@@ -2,7 +2,7 @@
  * rofi
  *
  * MIT/X11 License
- * Copyright © 2013-2021 Qball Cow <qball@gmpclient.org>
+ * Copyright © 2013-2022 Qball Cow <qball@gmpclient.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -54,11 +54,11 @@
 #include "settings.h"
 #include "timings.h"
 
-#include "dialogs/dialogs.h"
 #include "display.h"
 #include "helper-theme.h"
 #include "helper.h"
 #include "mode.h"
+#include "modes/modes.h"
 #include "xcb-internal.h"
 #include "xrmoptions.h"
 
@@ -111,6 +111,9 @@ struct {
   workarea mon;
   /** timeout for reloading */
   guint idle_timeout;
+  /** timeout for reloading */
+  guint refilter_timeout;
+  guint refilter_timeout_count;
   /** timeout handling */
   guint user_timeout;
   /** debug counter for redraws */
@@ -130,6 +133,8 @@ struct {
     .flags = MENU_NORMAL,
     .views = G_QUEUE_INIT,
     .idle_timeout = 0,
+    .refilter_timeout = 0,
+    .refilter_timeout_count = 0,
     .user_timeout = 0,
     .count = 0L,
     .repaint_source = 0,
@@ -194,7 +199,7 @@ void rofi_capture_screenshot(void) {
   if (outp == NULL) {
     int index = 0;
     fpath = g_build_filename(xdg_pict_dir, filename, NULL);
-    while (g_file_test(fpath, G_FILE_TEST_EXISTS) && index < 99) {
+    while (g_file_test(fpath, G_FILE_TEST_EXISTS) && index < 99999) {
       g_free(fpath);
       g_free(filename);
       // Try the next index.
@@ -269,6 +274,7 @@ static gboolean rofi_view_repaint(G_GNUC_UNUSED void *data) {
     // Repaint the view (if needed).
     // After a resize the edit_pixmap surface might not contain anything
     // anymore. If we already re-painted, this does nothing.
+
     rofi_view_update(current_active_menu, FALSE);
     g_debug("expose event");
     TICK_N("Expose");
@@ -308,19 +314,8 @@ static const int loc_transtable[9] = {
 static void rofi_view_calculate_window_position(RofiViewState *state) {
   int location = rofi_theme_get_position(WIDGET(state->main_window), "location",
                                          loc_transtable[config.location]);
-  int anchor = location;
-  if (!listview_get_fixed_num_lines(state->list_view)) {
-    anchor = location;
-    if (location == WL_CENTER) {
-      anchor = WL_NORTH;
-    } else if (location == WL_EAST) {
-      anchor = WL_NORTH_EAST;
-    } else if (location == WL_WEST) {
-      anchor = WL_NORTH_WEST;
-    }
-  }
-  anchor =
-      rofi_theme_get_position(WIDGET(state->main_window), "anchor", anchor);
+  int anchor =
+      rofi_theme_get_position(WIDGET(state->main_window), "anchor", location);
 
   if (CacheState.fullscreen) {
     state->x = CacheState.mon.x;
@@ -432,14 +427,38 @@ static void rofi_view_window_update_size(RofiViewState *state) {
   widget_resize(WIDGET(state->main_window), state->width, state->height);
 }
 
+extern GList *list_of_warning_msgs;
 static void rofi_view_reload_message_bar(RofiViewState *state) {
   if (state->mesg_box == NULL) {
     return;
   }
   char *msg = mode_get_message(state->sw);
-  if (msg) {
-    textbox_text(state->mesg_tb, msg);
+  if (msg || list_of_warning_msgs) {
+    /** we want to popin warning here. */
+
+    GString *emesg = g_string_new(msg);
+    if (list_of_warning_msgs) {
+      if (msg) {
+        g_string_append_c(emesg, '\n');
+      }
+      g_string_append(
+          emesg, "The following warnings were detected when starting rofi:\n");
+      GList *iter = g_list_first(list_of_warning_msgs);
+      int index = 0;
+      for (; iter != NULL && index < 2; iter = g_list_next(iter)) {
+        GString *msg = (GString *)(iter->data);
+        g_string_append(emesg, "\n\n");
+        g_string_append(emesg, msg->str);
+        index++;
+      }
+      if (g_list_length(iter) > 1) {
+        g_string_append_printf(emesg, "\nThere are <b>%u</b> more errors.",
+                               g_list_length(iter) - 1);
+      }
+    }
+    textbox_text(state->mesg_tb, emesg->str);
     widget_enable(WIDGET(state->mesg_box));
+    g_string_free(emesg, TRUE);
     g_free(msg);
   } else {
     widget_disable(WIDGET(state->mesg_box));
@@ -448,6 +467,13 @@ static void rofi_view_reload_message_bar(RofiViewState *state) {
 
 static gboolean rofi_view_reload_idle(G_GNUC_UNUSED gpointer data) {
   if (current_active_menu) {
+    // For UI update on this.
+    if (current_active_menu->tb_total_rows) {
+      char *r =
+          g_strdup_printf("%u", mode_get_num_entries(current_active_menu->sw));
+      textbox_text(current_active_menu->tb_total_rows, r);
+      g_free(r);
+    }
     current_active_menu->reload = TRUE;
     current_active_menu->refilter = TRUE;
     rofi_view_queue_redraw();
@@ -582,8 +608,8 @@ void rofi_view_free(RofiViewState *state) {
   g_free(state->distance);
   // Free the switcher boxes.
   // When state is free'ed we should no longer need these.
-  g_free(state->modi);
-  state->num_modi = 0;
+  g_free(state->modes);
+  state->num_modes = 0;
   g_free(state);
 }
 
@@ -1005,7 +1031,34 @@ inline static void rofi_view_nav_last(RofiViewState *state) {
   // state->selected = state->filtered_lines - 1;
   listview_set_selected(state->list_view, -1);
 }
+static void selection_changed_callback(G_GNUC_UNUSED listview *lv,
+                                       unsigned int index, void *udata) {
+  RofiViewState *state = (RofiViewState *)udata;
+  if (state->tb_current_entry) {
+    if (index < state->filtered_lines) {
+      int fstate = 0;
+      char *text = mode_get_display_value(state->sw, state->line_map[index],
+                                          &fstate, NULL, TRUE);
+      textbox_text(state->tb_current_entry, text);
+      g_free(text);
 
+    } else {
+      textbox_text(state->tb_current_entry, "");
+    }
+  }
+  if (state->icon_current_entry) {
+    if (index < state->filtered_lines) {
+      int icon_height =
+          widget_get_desired_height(WIDGET(state->icon_current_entry),
+                                    WIDGET(state->icon_current_entry)->w);
+      cairo_surface_t *icon =
+          mode_get_icon(state->sw, state->line_map[index], icon_height);
+      icon_set_surface(state->icon_current_entry, icon);
+    } else {
+      icon_set_surface(state->icon_current_entry, NULL);
+    }
+  }
+}
 static void update_callback(textbox *t, icon *ico, unsigned int index,
                             void *udata, TextBoxFontType *type, gboolean full) {
   RofiViewState *state = (RofiViewState *)udata;
@@ -1015,39 +1068,43 @@ static void update_callback(textbox *t, icon *ico, unsigned int index,
     char *text = mode_get_display_value(state->sw, state->line_map[index],
                                         &fstate, &add_list, TRUE);
     (*type) |= fstate;
-    // TODO needed for markup.
-    textbox_font(t, *type);
-    // Move into list view.
-    textbox_text(t, text);
-    PangoAttrList *list = textbox_get_pango_attributes(t);
-    if (list != NULL) {
-      pango_attr_list_ref(list);
-    } else {
-      list = pango_attr_list_new();
-    }
+
     if (ico) {
       int icon_height = widget_get_desired_height(WIDGET(ico), WIDGET(ico)->w);
       cairo_surface_t *icon =
           mode_get_icon(state->sw, state->line_map[index], icon_height);
       icon_set_surface(ico, icon);
     }
+    if (t) {
+      // TODO needed for markup.
+      textbox_font(t, *type);
+      // Move into list view.
+      textbox_text(t, text);
+      PangoAttrList *list = textbox_get_pango_attributes(t);
+      if (list != NULL) {
+        pango_attr_list_ref(list);
+      } else {
+        list = pango_attr_list_new();
+      }
 
-    if (state->tokens) {
-      RofiHighlightColorStyle th = {ROFI_HL_BOLD | ROFI_HL_UNDERLINE,
-                                    {0.0, 0.0, 0.0, 0.0}};
-      th = rofi_theme_get_highlight(WIDGET(t), "highlight", th);
-      helper_token_match_get_pango_attr(th, state->tokens,
-                                        textbox_get_visible_text(t), list);
+      if (state->tokens) {
+        RofiHighlightColorStyle th = {ROFI_HL_BOLD | ROFI_HL_UNDERLINE,
+                                      {0.0, 0.0, 0.0, 0.0}};
+        th = rofi_theme_get_highlight(WIDGET(t), "highlight", th);
+        helper_token_match_get_pango_attr(th, state->tokens,
+                                          textbox_get_visible_text(t), list);
+      }
+      for (GList *iter = g_list_first(add_list); iter != NULL;
+           iter = g_list_next(iter)) {
+        pango_attr_list_insert(list, (PangoAttribute *)(iter->data));
+      }
+      textbox_set_pango_attributes(t, list);
+      pango_attr_list_unref(list);
     }
-    for (GList *iter = g_list_first(add_list); iter != NULL;
-         iter = g_list_next(iter)) {
-      pango_attr_list_insert(list, (PangoAttribute *)(iter->data));
-    }
-    textbox_set_pango_attributes(t, list);
-    pango_attr_list_unref(list);
     g_list_free(add_list);
     g_free(text);
   } else {
+    // Never called.
     int fstate = 0;
     mode_get_display_value(state->sw, state->line_map[index], &fstate, NULL,
                            FALSE);
@@ -1103,9 +1160,11 @@ static void _rofi_view_reload_row(RofiViewState *state) {
   rofi_view_reload_message_bar(state);
 }
 
-static void rofi_view_refilter(RofiViewState *state) {
+static gboolean rofi_view_refilter_real(RofiViewState *state) {
+  CacheState.refilter_timeout = 0;
+  CacheState.refilter_timeout_count = 0;
   if (state->sw == NULL) {
-    return;
+    return G_SOURCE_REMOVE;
   }
   TICK_N("Filter start");
   if (state->reload) {
@@ -1119,6 +1178,8 @@ static void rofi_view_refilter(RofiViewState *state) {
   }
   TICK_N("Filter tokenize");
   if (state->text && strlen(state->text->text) > 0) {
+
+    listview_set_filtered(state->list_view, TRUE);
     unsigned int j = 0;
     gchar *pattern = mode_preprocess_input(state->sw, state->text->text);
     glong plen = pattern ? g_utf8_strlen(pattern, -1) : 0;
@@ -1131,6 +1192,9 @@ static void rofi_view_refilter(RofiViewState *state) {
      * speedup of the whole function.
      */
     unsigned int nt = MAX(1, state->num_lines / 500);
+    // Limit the number of jobs, it could cause stack overflow if we don´t
+    // limit.
+    nt = MIN(nt, config.threads * 4);
     thread_state_view states[nt];
     GCond cond;
     GMutex mutex;
@@ -1181,6 +1245,7 @@ static void rofi_view_refilter(RofiViewState *state) {
     state->filtered_lines = j;
     g_free(pattern);
   } else {
+    listview_set_filtered(state->list_view, FALSE);
     for (unsigned int i = 0; i < state->num_lines; i++) {
       state->line_map[i] = i;
     }
@@ -1220,6 +1285,33 @@ static void rofi_view_refilter(RofiViewState *state) {
   TICK_N("Filter resize window based on window ");
   state->refilter = FALSE;
   TICK_N("Filter done");
+  rofi_view_update(state, TRUE);
+  return G_SOURCE_REMOVE;
+}
+static void rofi_view_refilter(RofiViewState *state) {
+  CacheState.refilter_timeout_count++;
+  if (CacheState.refilter_timeout != 0) {
+
+    g_source_remove(CacheState.refilter_timeout);
+    CacheState.refilter_timeout = 0;
+  }
+  if (state->num_lines > config.refilter_timeout_limit &&
+      CacheState.refilter_timeout_count < 25 && state->text &&
+      strlen(state->text->text) > 0) {
+    CacheState.refilter_timeout =
+        g_timeout_add(200, (GSourceFunc)rofi_view_refilter_real, state);
+  } else {
+    rofi_view_refilter_real(state);
+  }
+}
+static void rofi_view_refilter_force(RofiViewState *state) {
+  if (CacheState.refilter_timeout != 0) {
+    g_source_remove(CacheState.refilter_timeout);
+    CacheState.refilter_timeout = 0;
+  }
+  if (state->refilter) {
+    rofi_view_refilter_real(state);
+  }
 }
 /**
  * @param state The Menu Handle
@@ -1362,6 +1454,12 @@ static void rofi_view_trigger_global_action(KeyBindingAction action) {
     state->retv = MENU_CANCEL;
     state->quit = TRUE;
     break;
+  case ELEMENT_NEXT:
+    listview_nav_next(state->list_view);
+    break;
+  case ELEMENT_PREV:
+    listview_nav_prev(state->list_view);
+    break;
   case ROW_UP:
     listview_nav_up(state->list_view);
     break;
@@ -1427,6 +1525,7 @@ static void rofi_view_trigger_global_action(KeyBindingAction action) {
     break;
   }
   case ACCEPT_ALT: {
+    rofi_view_refilter_force(state);
     unsigned int selected = listview_get_selected(state->list_view);
     state->selected_line = UINT32_MAX;
     if (selected < state->filtered_lines) {
@@ -1441,18 +1540,21 @@ static void rofi_view_trigger_global_action(KeyBindingAction action) {
     break;
   }
   case ACCEPT_CUSTOM: {
+    rofi_view_refilter_force(state);
     state->selected_line = UINT32_MAX;
     state->retv = MENU_CUSTOM_INPUT;
     state->quit = TRUE;
     break;
   }
   case ACCEPT_CUSTOM_ALT: {
+    rofi_view_refilter_force(state);
     state->selected_line = UINT32_MAX;
     state->retv = MENU_CUSTOM_INPUT | MENU_CUSTOM_ACTION;
     state->quit = TRUE;
     break;
   }
   case ACCEPT_ENTRY: {
+    rofi_view_refilter_force(state);
     // If a valid item is selected, return that..
     unsigned int selected = listview_get_selected(state->list_view);
     state->selected_line = UINT32_MAX;
@@ -1627,6 +1729,7 @@ void rofi_view_maybe_update(RofiViewState *state) {
     rofi_view_refilter(state);
   }
   rofi_view_update(state, TRUE);
+  return;
 }
 
 /**
@@ -1729,17 +1832,17 @@ static WidgetTriggerActionResult textbox_button_trigger_action(
   }
   return WIDGET_TRIGGER_ACTION_RESULT_IGNORED;
 }
-static WidgetTriggerActionResult textbox_sidebar_modi_trigger_action(
+static WidgetTriggerActionResult textbox_sidebar_modes_trigger_action(
     widget *wid, MouseBindingMouseDefaultAction action, G_GNUC_UNUSED gint x,
     G_GNUC_UNUSED gint y, G_GNUC_UNUSED void *user_data) {
   RofiViewState *state = (RofiViewState *)user_data;
   unsigned int i;
-  for (i = 0; i < state->num_modi; i++) {
-    if (WIDGET(state->modi[i]) == wid) {
+  for (i = 0; i < state->num_modes; i++) {
+    if (WIDGET(state->modes[i]) == wid) {
       break;
     }
   }
-  if (i == state->num_modi) {
+  if (i == state->num_modes) {
     return WIDGET_TRIGGER_ACTION_RESULT_IGNORED;
   }
 
@@ -1824,6 +1927,16 @@ static void rofi_view_add_widget(RofiViewState *state, widget *parent_widget,
                        TB_AUTOWIDTH | TB_AUTOHEIGHT, NORMAL, "", 0, 0);
     box_add((box *)parent_widget, WIDGET(state->tb_filtered_rows), FALSE);
     defaults = NULL;
+  } else if (strcmp(name, "textbox-current-entry") == 0) {
+    state->tb_current_entry =
+        textbox_create(parent_widget, WIDGET_TYPE_TEXTBOX_TEXT, name,
+                       TB_MARKUP | TB_AUTOHEIGHT, NORMAL, "", 0, 0);
+    box_add((box *)parent_widget, WIDGET(state->tb_current_entry), FALSE);
+    defaults = NULL;
+  } else if (strcmp(name, "icon-current-entry") == 0) {
+    state->icon_current_entry = icon_create(parent_widget, name);
+    box_add((box *)parent_widget, WIDGET(state->icon_current_entry), FALSE);
+    defaults = NULL;
   }
   /**
    * CASE INDICATOR
@@ -1882,11 +1995,9 @@ static void rofi_view_add_widget(RofiViewState *state, widget *parent_widget,
     }
     state->list_view = listview_create(parent_widget, name, update_callback,
                                        state, config.element_height, 0);
+    listview_set_selection_changed_callback(
+        state->list_view, selection_changed_callback, (void *)state);
     box_add((box *)parent_widget, WIDGET(state->list_view), TRUE);
-    // Set configuration
-    listview_set_multi_select(state->list_view,
-                              (state->menu_flags & MENU_INDICATOR) ==
-                                  MENU_INDICATOR);
     listview_set_scroll_type(state->list_view, config.scroll_method);
     listview_set_mouse_activated_cb(
         state->list_view, rofi_view_listview_mouse_activated_cb, state);
@@ -1907,17 +2018,17 @@ static void rofi_view_add_widget(RofiViewState *state, widget *parent_widget,
     state->sidebar_bar =
         box_create(parent_widget, name, ROFI_ORIENTATION_HORIZONTAL);
     box_add((box *)parent_widget, WIDGET(state->sidebar_bar), FALSE);
-    state->num_modi = rofi_get_num_enabled_modi();
-    state->modi = g_malloc0(state->num_modi * sizeof(textbox *));
-    for (unsigned int j = 0; j < state->num_modi; j++) {
+    state->num_modes = rofi_get_num_enabled_modes();
+    state->modes = g_malloc0(state->num_modes * sizeof(textbox *));
+    for (unsigned int j = 0; j < state->num_modes; j++) {
       const Mode *mode = rofi_get_mode(j);
-      state->modi[j] = textbox_create(
+      state->modes[j] = textbox_create(
           WIDGET(state->sidebar_bar), WIDGET_TYPE_MODE_SWITCHER, "button",
           TB_AUTOHEIGHT, (mode == state->sw) ? HIGHLIGHT : NORMAL,
           mode_get_display_name(mode), 0.5, 0.5);
-      box_add(state->sidebar_bar, WIDGET(state->modi[j]), TRUE);
+      box_add(state->sidebar_bar, WIDGET(state->modes[j]), TRUE);
       widget_set_trigger_action_handler(
-          WIDGET(state->modi[j]), textbox_sidebar_modi_trigger_action, state);
+          WIDGET(state->modes[j]), textbox_sidebar_modes_trigger_action, state);
     }
   } else if (g_ascii_strcasecmp(name, "overlay") == 0) {
     state->overlay = textbox_create(
@@ -2130,6 +2241,10 @@ void rofi_view_cleanup() {
     g_source_remove(CacheState.idle_timeout);
     CacheState.idle_timeout = 0;
   }
+  if (CacheState.refilter_timeout > 0) {
+    g_source_remove(CacheState.refilter_timeout);
+    CacheState.refilter_timeout = 0;
+  }
   if (CacheState.user_timeout > 0) {
     g_source_remove(CacheState.user_timeout);
     CacheState.user_timeout = 0;
@@ -2241,9 +2356,10 @@ void rofi_view_switch_mode(RofiViewState *state, Mode *mode) {
     rofi_view_set_window_title("rofi");
   }
   if (state->sidebar_bar) {
-    for (unsigned int j = 0; j < state->num_modi; j++) {
+    for (unsigned int j = 0; j < state->num_modes; j++) {
       const Mode *tb_mode = rofi_get_mode(j);
-      textbox_font(state->modi[j], (tb_mode == state->sw) ? HIGHLIGHT : NORMAL);
+      textbox_font(state->modes[j],
+                   (tb_mode == state->sw) ? HIGHLIGHT : NORMAL);
     }
   }
   rofi_view_restart(state);
